#ifndef __LINUX_KY_LPC_MULTIPLEXER_H
#define __LINUX_KY_LPC_MULTIPLEXER_H

#include <linux/types.h>
#include <linux/list.h>
#include <linux/spinlock.h>
#include <linux/refcount.h>

#include <linux/slab.h>
#include <linux/io.h>
#include <linux/device.h>

struct device;

extern struct list_head ky_lpcmux_list;

#define KY_LPCMUX_MAPSIZE   (4)

struct ky_lpc_mux {
	struct list_head    list;
	struct device      *dev;

	refcount_t          refcnt;
	spinlock_t          lock;
	unsigned long       lock_flags;

	phys_addr_t     addr_phy;
	int             addr_size;
	void __iomem   *addr_base;

	const void     *mux_path;
	int (*mux_switch) (struct ky_lpc_mux *mux, const void *mux_path);

	int pmap_size;
	int pmap_len;
	struct {
		const void *path;
		int   off;
		u64   val;
	}   path_map[0];
};

extern int
ky_lpcmux_switch32(struct ky_lpc_mux *mux, const void *mux_path);

extern struct ky_lpc_mux *
ky_lpcmux_alloc(int map_num, phys_addr_t addr_phy, int addr_size,
			int (* _switch) (struct ky_lpc_mux *, const void *));

extern void
ky_lpcmux_free(struct ky_lpc_mux *mux);

static inline struct ky_lpc_mux *
ky_lpcmux_find2(phys_addr_t addr_phy)
{
	struct list_head *_list;
	struct ky_lpc_mux *mux = NULL;

	list_for_each(_list, &ky_lpcmux_list) {
		struct ky_lpc_mux *mux_try = container_of(_list, struct ky_lpc_mux, list);
		if (mux_try->addr_phy == addr_phy) {
			mux = mux_try;
			break;
		}
	}

	return mux;
}

static inline struct ky_lpc_mux *
ky_lpcmux_find(int map_num, phys_addr_t addr_phy, int addr_size,
			int (* _switch) (struct ky_lpc_mux *, const void *))
{
	struct ky_lpc_mux *mux = NULL;

	mux = ky_lpcmux_find2(addr_phy);
	if (mux == NULL) {
		mux = ky_lpcmux_alloc(map_num, addr_phy, addr_size, _switch);
	} else {
		refcount_inc(&mux->refcnt);
	}

	dev_dbg(mux->dev, "ky_lpcmux_find(91): mux=%p, refcnt=%d\n",
			mux, (int)refcount_read(&mux->refcnt));
	return mux;
}

static inline int
ky_lpcmux_muxadd(struct ky_lpc_mux *mux, const void *mux_path, int reg_off, u64 reg_val)
{
	int  pmap_idx;
	unsigned long lock_flags;

	dev_dbg(mux->dev, "ky_lpcmux_muxadd(00): mux=%p, new path=%p ..\n",
			mux, mux_path);

	spin_lock_irqsave(&mux->lock, lock_flags);

	for (pmap_idx = 0; pmap_idx <  mux->pmap_len; pmap_idx++) {
		if (mux_path == mux->path_map[pmap_idx].path)
			break;
	}
	if ((pmap_idx >= mux->pmap_len) && (mux->pmap_len >= mux->pmap_size))
		goto err_no_map;

	if (pmap_idx >= mux->pmap_len)
		pmap_idx = mux->pmap_len++;
	mux->path_map[pmap_idx].path = mux_path;
	mux->path_map[pmap_idx].off  = reg_off;
	mux->path_map[pmap_idx].val  = reg_val;

	spin_unlock_irqrestore(&mux->lock, lock_flags);

	dev_dbg(mux->dev, "ky_lpcmux_muxadd(91): mux=%p, new path=%p as pmap_new=%d/%d\n",
			mux, mux_path, pmap_idx, mux->pmap_len);
	return 0;

err_no_map:
	spin_unlock_irqrestore(&mux->lock, lock_flags);
	return -ENOSPC;
}

static inline int
ky_lpcmux_muxdel(struct ky_lpc_mux *mux, const void *mux_path)
{
	int  pmap_idx, pmap_del;
	unsigned long lock_flags;

	for (pmap_idx = 0; pmap_idx <  mux->pmap_len; pmap_idx++) {
		if (mux_path == mux->path_map[pmap_idx].path)
			break;
	}
	if (pmap_idx >= mux->pmap_len)
		goto err_no_map;

	spin_lock_irqsave(&mux->lock, lock_flags);

	pmap_del = pmap_idx;
	mux->pmap_len--;
	for (; pmap_idx <  mux->pmap_len; pmap_idx++) {
		dev_dbg(mux->dev, "ky_lpcmux_muxdel(61): mux=%p, move path pmap_idx(%d => %d)\n",
				mux, pmap_idx+1, pmap_idx);
		memmove(&mux->path_map[pmap_idx], &mux->path_map[pmap_idx+1],
				sizeof(mux->path_map[pmap_idx]));
	}
	dev_dbg(mux->dev, "ky_lpcmux_muxdel(65): mux=%p, clean path pmap_idx=%d\n",
			mux, pmap_idx);
	mux->path_map[pmap_idx].path = NULL;
	mux->path_map[pmap_idx].off  = 0;
	mux->path_map[pmap_idx].val  = 0;

	spin_unlock_irqrestore(&mux->lock, lock_flags);

	dev_dbg(mux->dev, "ky_lpcmux_muxdel(91): mux=%p, del path=%p as pmap_del=%d/%d\n",
			mux, mux_path, pmap_del, mux->pmap_len+1);
	return 0;

err_no_map:
	return -ENXIO;
}

static inline int
ky_lpcmux_lock(struct ky_lpc_mux *mux, const void *mux_path)
{
	if (!mux)
		return 0;

	spin_lock_irqsave(&mux->lock, mux->lock_flags);
	if (mux_path != mux->mux_path && mux->mux_switch) {
		mux->mux_switch(mux, mux_path);
		mux->mux_path = mux_path;
	}
	return 0;
}

static inline int
ky_lpcmux_unlock(struct ky_lpc_mux *mux)
{
	if (!mux)
		return 0;

	spin_unlock_irqrestore(&mux->lock, mux->lock_flags);
	return 0;
}

struct kylpc_cdev_conf {
	struct device  *dev;
	int     ver_fw;

	/* device generial resources */
	int             phy_irq;
	unsigned long   io_base;
	long            io_off;
	unsigned long   io_len;

	/* device init arguments by FW spec */
	int     num_init;
	struct {
		int    type;
		u64    off;
		u64    val;
	} init_regs[4];

	/* device switch arguments by FW spec */
	int     flag_switch;
	struct {
		int    type;
		u64    off;
		u64    val;
	} switch_regs;

	/* device IO behaviour */
	unsigned int    io_endian : 1,
			/* 0: little endian */
				io_size : 8;
			/* 0x01:  1 byte mode support */
			/* 0x02:  2 bytes mode support */
			/* 0x04:  4 bytes mode support */
			/* 0x08:  8 bytes mode support */
};

extern int  kylpc_cdev_conf_init(struct kylpc_cdev_conf *pconf);
extern int  kylpc_cdev_conf_load(struct kylpc_cdev_conf *pconf,
				struct device *dev);

#endif/* __LINUX_KY_LPC_MULTIPLEXER_H */

