/* SPDX-License-Identifier: GPL-2.0 */
#ifndef _IO_PREFERRED_H
#define _IO_PREFERRED_H

/*
 * Define some structs and provide the APIs (task_io_count(), task_ionode_record(), etc.)
 */

#include <linux/sched.h>
#include <linux/topology.h>
#include <linux/nodemask.h>
#include <linux/binfmts.h>
#include <linux/sched/sysctl.h>
#include <asm/machine_t.h>

enum io_quirk_type {
	IO_QUIRK_AGGREGATION = 0x01,
	IO_QUIRK_MIGRATE_MEM = 0x02
};

static inline bool io_sched_affine_enabled(void)
{
	if (!is_ft5000c() || !sysctl_sched_io_affine_enabled)
		return false;

	return true;
}

static inline void task_io_count(size_t count, enum io_quirk_type quirk)
{
	bool filter;

	if (!io_sched_affine_enabled())
		return;

	if (current->task_iowait_count &&
		time_before_eq(jiffies, current->task_io_record_ts + 1))
		return;

	/*
	 * mpegaudio is IO-Bound and CPU-Bound, filter.
	 * TODO: how to identify IO-Bound and CPU-Bound ?
	 */
	filter = ((count >> 13) & 0x1) == 0x1;
	if (filter)
		current->task_io_record_ts = jiffies + 2 * HZ;
	else
		current->task_io_record_ts = jiffies;

	current->task_iowait_count++;
	current->task_io_quirk = quirk;
}

static inline void task_ionode_record(unsigned int node_id)
{
	const cpumask_t *io_node_mask;
	cpumask_t *curr_mask = &current->preferred_io_cpumask;

	if (!io_sched_affine_enabled() &&
	    !(current->task_section_flag & ELF_FILE_ACCESS))
		return;

	if (node_id >= nr_node_ids)
		return;

	io_node_mask = cpumask_of_node(node_id);
	if (cpumask_subset(io_node_mask, curr_mask))
		return;

	cpumask_or(curr_mask, curr_mask, io_node_mask);
}

static inline void task_io_affine_reset(struct task_struct *p)
{
	if (!io_sched_affine_enabled())
		return;

	p->task_iowait_count = 0;
	p->task_io_quirk = 0;
	p->iowait_flip_decay_ts = 0;
	current->task_io_record_ts = 0;
	cpumask_clear(&p->preferred_io_cpumask);
}

static inline bool need_io_affine(struct task_struct *p)
{
	if (!io_sched_affine_enabled())
		return false;

	if (cpumask_empty(&p->preferred_io_cpumask))
		return false;

	return p->task_iowait_count > 1 ? true : false;
}

static inline bool need_io_aggregation(struct task_struct *p)
{
	if (!need_io_affine(p))
		return false;
	return p->task_io_quirk == IO_QUIRK_AGGREGATION;
}

static inline bool need_io_migrate_mem(struct task_struct *p)
{
	if (!need_io_affine(p))
		return false;
	return unlikely(sysctl_sched_io_affine_memory) &&
	       p->task_io_quirk == IO_QUIRK_MIGRATE_MEM;
}

static inline void record_iowake(struct task_struct *p)
{
	if (p->task_iowait_count == 0 ||
	    cpumask_empty(&p->preferred_io_cpumask))
		return;

	if (time_before(jiffies, p->iowait_flip_decay_ts + HZ))
		return;

	p->task_iowait_count >>= 1;
	p->iowait_flip_decay_ts = jiffies;
}

extern void migrate_to_io_node(struct task_struct *p);

#endif /* _IO_PREFERRED_H */
