#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import argparse
import subprocess
import os
import re
import time
import git
import openpyxl
import xlrd
from openpyxl.styles import Alignment, Border, Side

thin_border = Border(
    left=Side(style='thin'),
    right=Side(style='thin'),
    top=Side(style='thin'),
    bottom=Side(style='thin')
)

trees_url = {
    "klinux": "ssh://kernel-bot@gerrit.kylin.com:29418/klinux.git",
    "klinux-4.19": "ssh://kernel-bot@gerrit.kylin.com:29418/klinux-4.19.git",
    "kfocal": "ssh://kernel-bot@gerrit.kylin.com:29418/kfocal.git"
}

firm_list = ["HYGON", "PHYTIUM", "Sunway", "KYLIN"]

result = None


# 执行命令函数
def cmd_call(cmd):
    print(cmd)
    res = subprocess.Popen(cmd,
                           shell=True,
                           stdin=subprocess.PIPE,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE)
    output, errors = res.communicate()
    try:
        op = output.decode('utf-8').splitlines()
    except UnicodeDecodeError:
        op = output.decode('ISO-8859-1').splitlines()
    err = errors.decode('utf-8').splitlines()
    res.wait()
    res.stdout.close()
    state = res.returncode
    return state, op, err


def get_cur_branch():
    state, output, error = cmd_call("git branch;")
    for branch in output:
        if '*' in branch:
            return branch.split(' ')[-1]
    return None


def key_begin_check(firm_name, branches):
    title_dict = {}
    data_list = []
    # 获取所有符合条件的标题
    for branch in branches:
        status, op, err = cmd_call("git log {} --pretty=oneline --grep='^{}'".format(branch, firm_name))
        if len(op) < 1:
            print("找不到符合条件的数据！")
        for msg in op:
            kylin_commit_id = msg.split(" ")[0].strip()
            commit_title = msg[41:].strip()
            if title_dict.get(commit_title):
                title_dict[commit_title][branch + "_commit_id"] = kylin_commit_id
            else:
                title_dict[commit_title] = {branch + "_commit_id": kylin_commit_id}
    # 依次获取标题的版本信息：
    for title, title_msg in title_dict.items():
        data_dict = init_data_dict(branches)
        data_dict["title"] = title
        for branch in branches:
            branch_commit_id = title_msg.get(branch + "_commit_id")
            if branch_commit_id:
                get_commit_msg(branch, branch_commit_id, data_dict)
        data_list.append(data_dict)
    return data_list


def get_commit_msg(branch, branch_commit_id, data_dict):
    data_dict[branch + "_commit_flag"] = "Y"
    # 分支已合并，获取提交人及版本信息
    status, op, err = cmd_call(
        "git show %s |grep Signed-off-by |grep @kylinos.cn |tail -n 1 |awk -F ':' '{print $2}'" % (branch_commit_id))
    if op:
        data_dict[branch + "_commit_by"] = op[0].strip()
        status, op, err = cmd_call('git describe {}'.format(branch_commit_id))
        if op:
            data_dict[branch + "_version"] = op[0].strip()
    else:
        data_dict[branch + "_commit_flag"] = "找不到kylin提交人信息"


def init_data_dict(branches):
    data_dict = {
        "title": "",
        "is_up": "N",
        "up_commit_id": ""
    }
    for branch in branches:
        data_dict[branch + "_commit_flag"] = "N"
        data_dict[branch + "_version"] = ""
        data_dict[branch + "_commit_by"] = ""
    return data_dict


def check_master(title, data_dict):
    # 通过master分支确认是否为上游补丁
    status, op, err = cmd_call("git log master --pretty=oneline --grep='^{}'".format(title))
    if op:
        data_dict["is_up"] = "Y"
        data_dict["up_commit_id"] = op[0].split(" ")[0].strip()


def task_id_begin_check(task_id_list, branches):
    title_dict = {}
    data_list = []
    for task_id in task_id_list:
        for branch in branches:
            status, op, err = cmd_call("git log {} --grep='task: {}' --pretty=oneline".format(branch, task_id))
            if len(op) < 1:
                print("找不到符合条件的数据！")
            for msg in op:
                kylin_commit_id = msg.split(" ")[0].strip()
                commit_title = msg[41:].strip()
                if title_dict.get(commit_title):
                    title_dict[commit_title][branch + "_commit_id"] = kylin_commit_id
                else:
                    title_dict[commit_title] = {branch + "_commit_id": kylin_commit_id}

    cmd_call("git checkout master")
    for title, title_msg in title_dict.items():
        data_dict = init_data_dict(branches)
        data_dict["title"] = title
        if title.split(":")[0].strip() not in firm_list:
            # 去master分支检查是否为上游补丁
            check_master(title, data_dict)
        for branch in branches:
            branch_commit_id = title_msg.get(branch + "_commit_id")
            if branch_commit_id:
                get_commit_msg(branch, branch_commit_id, data_dict)
        data_list.append(data_dict)
    return data_list


def id_begin_check(up_commit_id_list, branches):
    title_dict = {}
    data_list = []
    for up_commit_id in up_commit_id_list:
        for branch in branches:
            status, op, err = cmd_call("git log {} --grep='Mainline: {}' --pretty=oneline".format(branch, up_commit_id))
            if len(op) < 1:
                print("找不到符合条件的数据！")
            for msg in op:
                kylin_commit_id = msg.split(" ")[0].strip()
                commit_title = msg[41:].strip()
                if title_dict.get(commit_title):
                    title_dict[commit_title][branch + "_commit_id"] = kylin_commit_id
                    title_dict[commit_title]["up_commit_id"] = up_commit_id
                else:
                    title_dict[commit_title] = {branch + "_commit_id": kylin_commit_id, "up_commit_id": up_commit_id}
    for title, title_msg in title_dict.items():
        data_dict = init_data_dict(branches)
        data_dict["title"] = title
        data_dict["is_up"] = "Y"
        data_dict["up_commit_id"] = title_msg.get("up_commit_id")
        for branch in branches:
            branch_commit_id = title_msg.get(branch + "_commit_id")
            if branch_commit_id:
                get_commit_msg(branch, branch_commit_id, data_dict)
        data_list.append(data_dict)
    return data_list


def get_up_commit_id(title):
    # 通过master分支确认标题是否为上游补丁，是，返回上游commit_id
    status, op, err = cmd_call("git log master --oneline --grep='{}'".format(title))
    for commit in op:
        commit_msg = commit.split(" ", 1)
        commit_id = commit_msg[0].strip()
        commit_title = commit_msg[1].strip()
        if commit_title == title:
            return commit_id


def format_title(title):
    # 处理标题
    s_p = re.compile(r'[^a-zA-Z0-9\s()\'<>=:|./~-]')
    new_title = re.sub(s_p, lambda x: '\\' + x.group(0), title)
    # 标题中存在\n,\t,\r的处理
    new_title = new_title.replace(r'\\', "\\\\\\")
    return new_title


def get_branch_data(src_title, des_title, branch_list, data_dict):
    for branch in branch_list:
        print("--------------当前检查分支为：{}-----------".format(branch))
        status, op, err = cmd_call('git log {} --oneline --grep="{}"'.format(branch, des_title))
        kylin_commit = ""
        # 检查标题和上游补丁信息
        for commit in op:
            commit_msg = commit.split(" ", 1)
            kylin_commit_id = commit_msg[0].strip()
            commit_title = commit_msg[1].strip()
            # 标题不匹配，截取厂商名。
            if commit_title != src_title:
                split_key_title = commit_title.split(":", 1)
                if len(split_key_title) > 1:
                    if split_key_title[1].strip() == src_title:
                        kylin_commit = kylin_commit_id
                        data_dict["补丁标题"] = commit_title
                        break
            else:
                kylin_commit = kylin_commit_id
                break
        # 分支已合并
        if kylin_commit:
            # 获取合入人
            status, op, err = cmd_call(
                "git show %s |grep Signed-off-by |grep @kylinos.cn |tail -n 1 |awk -F ':' '{print $2}'" % (
                    kylin_commit))
            if op:
                data_dict["合入人"] = op[0].strip()
            # 获取合入版本
            status, op, err = cmd_call('git describe {}'.format(kylin_commit))
            if op:
                data_dict[branch + "合入版本"] = op[0].strip()
        else:
            data_dict[branch + "合入版本"] = "N"
    print("----------当前补丁检查结果-----------")
    print(data_dict)
    return data_dict


def patch_check(patch_dir, branch_list):
    # 创建新的 Excel 工作簿
    wb = openpyxl.Workbook()
    # 获取当前活跃的工作表
    ws = wb.active
    # 定义标题行数据
    headers = ["提交时间", "补丁版本", "补丁说明", "补丁标题", "是否上游", "Upstream Commit ID"]
    for branch in branch_list:
        headers.append(branch + "合入版本")
    headers.append("合入人")
    headers.append("备注")
    # 插入标题行
    for col, header in enumerate(headers, 1):
        ws.cell(1, col, header)
        ws.cell(row=1, column=col).border = thin_border  # 设置边框
        ws.cell(row=1, column=col).alignment = Alignment(horizontal='left', vertical='center')

    # 组建数据行
    data_list = []
    # 依次获取某个目录下的patch文件，并提取文件中的标题。
    print("---------开始patch文件补丁合并检查----------")
    for root, dirs, files in os.walk(patch_dir):
        dirs.sort()
        for name in files:
            filename = os.path.join(root, name)
            if filename.endswith('.patch'):
                with open(filename) as f:
                    print("读取patch文件：{}".format(filename))
                    content = f.read()
                    subject_pattern = re.compile(r'Subject: \[.*?\](.*?)\n\n', re.DOTALL)
                    match = subject_pattern.search(content)

                    src_title = match.group(1).replace('\n', '').strip()
                    print("在patch文件中提取到的标题为：{}".format(src_title))
                    data_dict = dict()
                    data_dict["补丁标题"] = src_title
                    print("当前检查补丁标题：{}".format(src_title))
                    # 处理标题
                    des_title = format_title(src_title)
                    # 获取上游commit_id
                    up_commit_id = get_up_commit_id(des_title)
                    if up_commit_id:
                        data_dict["是否上游"] = "Y"
                        data_dict["Upstream Commit ID"] = up_commit_id
                    else:
                        data_dict["是否上游"] = "N"
                    # 获取分支合并情况
                    data_dict = get_branch_data(src_title, des_title, branch_list, data_dict)
                    data_list.append(data_dict)
    # 插入数据行
    for row, d in enumerate(data_list, 2):
        for col, header in enumerate(headers, 1):
            ws.cell(row=row, column=col).border = thin_border  # 设置边框
            ws.cell(row=row, column=col).alignment = Alignment(horizontal='left', vertical='center')
            if d.get(header):
                ws.cell(row, col, d[header])
            else:
                ws.cell(row, col, "")
    # 保存Excel文件
    result_dir = patch_dir + "/check_result"
    if not os.path.exists(result_dir):
        # 如果不存在则创建文件夹
        os.mkdir(result_dir)
    os.chdir(result_dir)
    timestamp = time.strftime("%Y%m%d%H%M%S")
    file_name = "patch_check_result_{}.xlsx".format(timestamp)
    file_path = str(result_dir) + '/' + file_name
    wb.save(file_name)
    print("------补丁检查完毕--------")
    print("检查结果为：{}".format(file_path))


def check_type(args):
    global result
    code_dir = args.code_dir
    branches = args.branches
    # 判断检查类型：
    if args.firm_name:
        firm_name = args.firm_name
        print("检查类型：厂商名:{}".format(firm_name))
        branch_list = git_check_update(branches, code_dir)
        result = key_begin_check(firm_name, branch_list)
    elif args.up_commit_id:
        up_commit_id = args.up_commit_id.split(",")
        print("检查类型：上游commit_id:{}".format(up_commit_id))
        branch_list = git_check_update(branches, code_dir)
        result = id_begin_check(up_commit_id, branch_list)
    elif args.task_id:
        task_id = args.task_id.split(",")
        print("检查类型：禅道任务ID:{}".format(task_id))
        branch_list = git_check_update(branches, code_dir)
        result = task_id_begin_check(task_id, branch_list)
    elif args.patch_dir:
        patch_dir = args.patch_dir
        print("检查类型：补丁文件目录:{}".format(patch_dir))
        branch_list = git_check_update(branches, code_dir)
        update_branch("master")
        patch_check(patch_dir, branch_list)
    else:
        print("中断脚本执行：无效的检查类型")
        exit()
    # xlwt实现方式
    if result:
        # 根据结果输出到excel文件
        import xlwt
        # 创建Excel文件对象、表格页sheet
        workbook = xlwt.Workbook(encoding='utf-8')
        worksheet = workbook.add_sheet('补丁检查结果')

        # sheet页中写入数据
        table_head = ['补丁标题', '是否上游', '上游commit_id']
        k_list = ["title", "is_up", "up_commit_id"]
        for branch in branch_list:
            table_head.append("{}是否合入".format(branch))
            table_head.append("{}合入版本".format(branch))
            table_head.append("{}合入提交人".format(branch))
            k_list.append(branch + "_commit_flag")
            k_list.append(branch + "_version")
            k_list.append(branch + "_commit_by")
        for index in range(len(table_head)):
            worksheet.write(0, index, label=table_head[index])  # 先添加一行表头

        for col_index in range(len(result)):
            for row_index in range(len(result[col_index])):
                key = k_list[row_index]
                worksheet.write(col_index + 1, row_index, label=result[col_index].get(key))  # 添加每行数据

        # 保存Excel文件
        result_dir = code_dir + "/check_result"
        if not os.path.exists(result_dir):
            # 如果不存在则创建文件夹
            os.mkdir(result_dir)
        os.chdir(result_dir)
        timestamp = time.strftime("%Y%m%d%H%M%S")
        file_name = "check_result_{}.xlsx".format(timestamp)
        file_path = str(result_dir) + '/' + file_name
        workbook.save(file_name)
        print("------补丁检查完毕--------")
        print("检查结果为：{}".format(file_path))
        print(file_path)


def git_check_update(branches, code_dir):
    if os.path.exists(code_dir):
        try:
            git.Repo(code_dir)
            print('need to update code')
            os.chdir(code_dir)
            branch_list = update_branch(branches)
            return branch_list
        except git.InvalidGitRepositoryError:
            print("目录：{} 不是一个git仓，请核对。".format(code_dir))
            exit()
    else:
        print("目录不存在：{} 请核对。".format(code_dir))
        exit()


def update_branch(branches):
    branch_list = []
    # 判断是否携带了分支名
    if branches:
        branch_list = branches.split(",")
    else:
        # 获取当前分支名
        cur_branch = get_cur_branch()
        print("用户未携带分支，默认检查当前分支：{}".format(cur_branch))
        branch_list.append(cur_branch)
    # 更新代码
    for branch in branch_list:
        print("更新分支代码:{}".format(branch))
        # 依次更新分支代码
        cur_branch = get_cur_branch()
        if branch != cur_branch:
            state, op, err = cmd_call("git checkout {}".format(branch))
            if state != 0:
                print("-----{}分支切换失败，检查终止运行-----".format(branch))
                for err in err:
                    print(err)
                exit()
        state, op, err = cmd_call("git pull")
        if state != 0:
            print("-----{}分支更新失败，检查终止运行-----".format(branch))
            for err in err:
                print(err)
            exit()
    return branch_list


def main():
    parser = argparse.ArgumentParser(
        description='补丁合并检查脚本：支持厂商名，禅道任务id，上游commit_id，批量patch文件 四种方式')
    parser.add_argument('-d', '--code_dir', help='可选项，仓库目录，需是绝对路径。默认当前目录。如/data/klinux-4.19',
                        default=os.getcwd())
    parser.add_argument('-b', '--branches',
                        help='可选项，默认检查当前分支。检查分支名，支持多个，用英文逗号隔开，如kylinos-next, stable-52-sp3')
    parser.add_argument('-k', '--firm_name', help='可选项，厂商名，如HYGON')
    parser.add_argument('-t', '--task_id', help='可选项，禅道任务id，如73883')
    parser.add_argument('-i', '--up_commit_id',
                        help='可选项，上游commit_id，支持多个，用英文逗号隔开。如ef1491e791308317bb9851a0ad380c4a68b58d54')
    parser.add_argument('-p', '--patch_dir', help='可选项，patch文件目录，如/home/helen/test/patch')
    parser.set_defaults(func=check_type)
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()
